/**************************************************************************

Copyright [2012] [Software Engineering Group - Kiel University]

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

***************************************************************************/

var labelType, useGradients, nativeTextSupport, animate;

/**
	Set up Error Text and Navigator
*/

(function() {
  var ua = navigator.userAgent,
      iStuff = ua.match(/iPhone/i) || ua.match(/iPad/i),
      typeOfCanvas = typeof HTMLCanvasElement,
      nativeCanvasSupport = (typeOfCanvas == 'object' || typeOfCanvas == 'function'),
      textSupport = nativeCanvasSupport 
        && (typeof document.createElement('canvas').getContext('2d').fillText == 'function');
  //I'm setting this based on the fact that ExCanvas provides text support for IE
  //and that as of today iPhone/iPad current text support is lame
  labelType = (!nativeCanvasSupport || (textSupport && !iStuff))? 'Native' : 'HTML';
  nativeTextSupport = labelType == 'Native';
  useGradients = nativeCanvasSupport;
  animate = !(iStuff || !nativeCanvasSupport);
  
  // fix IE script crashes
  if(!window.console){
	  window.console={log:function(){}};
	  }
})();

var Log = {
  elem: false,
  write: function(text){
    if (!this.elem) 
      this.elem = document.getElementById('log');
    this.elem.innerHTML = text;
    this.elem.style.left = (500 - this.elem.offsetWidth / 2) + 'px';
  }
};


function GraphFlow(){
	/////////////////////////////////////////////
	//				VARIABLES				   //
	/////////////////////////////////////////////
	
	/** Overall size modifier: the width of node labels in pixels */
	var vertexLabelSize = 12;
	
	var visContainer = 'infovis';
	
	/** Colors */
	var nodeFillColor = [];
	nodeFillColor['nodeFamily']		= "#DEDEDE";
	
	var nodeStrokeColor = [];
	nodeStrokeColor['inputPort']   	  = "#AA0000";
	nodeStrokeColor['outputPort']	  = "#AA0000";
	nodeStrokeColor['repositoryPort'] = "#AA0000";
	nodeStrokeColor['nodeFamily'] 	  = "#4D4D4D";
	nodeStrokeColor['crossBox'] 	  = "#4D4D4D";
		
	var nodeColorFocus	  = "#0098BE",
		edgeColor 		  = "#114270",
		edgeColorFocus 	  = "#0098BE";
		
	/** Custom Node Edges **/
	var customEdgeMap = {
		'none' : 
		{
			'getOffset' : function(){return {'x' : 0, 'y' : -2}}
		},
		'inputPort' : 
			{
			'getOffset' : function(){return {'x' : -vertexLabelSize, 'y' : vertexLabelSize / 2}}
			},
		'outputPort' : 
			{
			'getOffset' : function(){return {'x' : vertexLabelSize, 'y' : vertexLabelSize / 2}}
			}
	};
	customEdgeMap['repositoryPort'] = customEdgeMap['outputPort'];
	
	/** Icon Images */
	var nodeIcons = [];
	
	/** GraphFlow Object */
	var fd;

	/** Nodes and edges that are highlighted when the mouse moves over them */
	var hover = null;
	var markedNode = null;

	/** selected node:	{id : the id of the selected node, 
						 from : the data.$type of the selected node,
						 label: the label of a selected edge}
						 
		When clicking a node or edge, this variable will contain node
		information, that is required for attaching/detaching edges.
		*/
	var selectedNode = null;
	
	/** This points to an an invisible dummy node which is
		 used for moving an edge with the mouse
	*/
	var mouseNode = null;

	/** true if the mouse is over a port/edge */
	var mouseOverNode = false;
	var mouseOverEdge = false;
	
	/** A hashtable, mapping mouse and node events to arrays of functions that
		are called when these events happen
		*/
	var listener = [];
	
	/** Can be changed to allow or prohibit node movement or deletion/creation */
	var readOnly = {'deleteCreate' 	: true,
					'move' 			: true};
	
	/** Grid dummy-node which is drawn over the graph if it is visible */
		
	var grid;
	
	/** Stores information about the canvas, such as its dimensions and its position.
		This information is needed to correctly draw the Grid and may come in handy in
		future functions.
		*/
	var navi = { 
		 // screen position
		 'centerX' : 0,
		 'centerY' : 0,
		 'isDragging' : false,
		 'isDraggingEdge' : false,
		 // how far a the screen is dragged / where a dragged node is touched
		 'dragX' : null,
		 'dragY' : null};
	
	var animation = {
		'enabled' : true,
		'duration' : 1000,
		'busyCanvas' : false
	};
	
	/////////////////////////////////////////////
	//				FUNCTIONS				   //
	/////////////////////////////////////////////

	
	/**
	 * Enables / Disables graph animation.
	 * 
	 * @param enabled - if true, animation is enabled
	 * @param duration - the duration of animations
	 */
	this.setAnimation = function(enabled, duration){
		
		animation.enabled = enabled;
		if(duration){
			animation.duration = duration;
		}
	}
	
	/**
	 * Measures the width of a text on the canvas
	 * 
	 * @param text the text that is measured
	 * @param font the size and font of the text
	 * @returns the measured width
	 */
	this.getTextWidth = function(text, font){
		
		/*
		// validate arguments
		if(!validArg("getTextWidth()", text, "string") ||
		   !validArg("getTextWidth()", font, "string")){
			return;
		}
		*/
			
		// get canvas context
		var ctx = fd.canvas.getCtx(),
			width = 0;
		
		// measure width
		ctx.save();
		ctx.font = font;
		width = ctx.measureText(text).width;
		ctx.restore();
		
		return width;
	}
	
	/**
	 * @returns the size modifier that scales the whole graph.
	 */
	this.getSizeModifier = function(){ 
		return vertexLabelSize;
	}
	
	/**
	 * Adds a node (that is NOT a Nodefamily) to the graph.
	 * 
	 * @param xPosition - the horizontal position of the node
	 * @param yPosition - the vertical position of the node
	 * @param node - an object which at least contains an 'id' and may have a 'data' field 
	 * @param type - the type of the defined custom node (basic types are: inputPort, nodeFamily, ...)
	 * @param forced - if true, does not throw a listener event
	 */
	this.addCustomNode = function(xPosition, yPosition, node, type, forced){
		
		// validate arguments
		var myName = "addCustomNode()";
		
		/*
		if(!validArg(myName, xPosition, "number") ||
		   !validArg(myName, yPosition, "number") ||
		   !validArg(myName, type, "string") ||
		   !validArgNode(myName, node) ||
		   !validArgID(myName, node.id)){
			return false;
		}
		*/
		
		var nodeType = 'Custom';
		
		if(node.data && node.data.$nodeType){
			nodeType = node.data.$nodeType;
		}
		
		// define node
		var newNode = {
				  "adjacencies": [], 
				  "data": {
					"$dim": vertexLabelSize,		
					"$type": type,
					"$icon": nodeIcons[nodeType],
					"$nodeType": nodeType,
					"$custom": true
				  }, 
				  "id": node.id, 
				  "name": node.name};
		
		// add colors if they exist
		var	strokeColor = nodeStrokeColor[type],
			nodeColor = nodeFillColor[type];
		if(strokeColor){ newNode.data.$color = strokeColor;}
		if(nodeColor){ newNode.data.$fillColor = nodeColor;}
		
		// (over-)write data if it was specified in node.data
		if(node.data){
			for(var field in node.data){
				newNode.data[field] = node.data[field];
			}
		}
		
		// apply changes immediately
		fd.graph.addNode(newNode);
		newNode = fd.graph.getNode(newNode.id);
		newNode.pos.setc(xPosition, yPosition);
		fd.plot();
		
		// call listener
		if(!forced){
			callListener("onCreateNode", [newNode]);
		}
		return newNode;
	}
	
	/**
	 * Zooms in or out, depending on the factor
	 * 
	 * @param factor -  the graph zooms out if it is smaller than 1
	 */
	this.zoom = function(factor){
		
		/*
		// check argument
		if(!validArg("zoom()", factor, "number")){
			return;
		}else if(factor < 0){
			callListener("onError", ["Error in function zoom(): Argument factor must be a positive number!"]);
		}
		*/
		
		// zoom
		var canvas = fd.canvas;
		canvas.scale(factor, factor);
		
		// show text only if it is big enough
		canvas.showLabels = (canvas.scaleOffsetX > canvas.labelThreshold);
		fd.plot();
	}
	
	/**
		Checks if the GraphFlow container has changed in size and
		updates the canvas area accordingly.
	*/
	this.updateCanvasSize = function(){
		var container = document.getElementById(visContainer);
		var widthOld = grid.data.$width,
			heightOld = grid.data.$height,
			widthNew = container.clientWidth,
			heightNew = container.clientHeight;
			
		// check if the dimensions changed
		if( widthOld != widthNew || heightOld != heightNew){
		
			// calculate translation parameters
			var scale = fd.canvas.scaleOffsetX,
				transX = fd.canvas.translateOffsetX + (widthOld - widthNew)/2,
				transY = fd.canvas.translateOffsetY + (heightOld - heightNew)/2;
			
			// resize canvas, scale and translate content to maintain its position
			fd.canvas.resize(container.clientWidth, container.clientHeight, true);
			fd.canvas.scale(scale, scale);
			fd.canvas.translate(transX / scale, transY / scale);
			
			// update grid dimensions
			grid.data.$width = widthNew;
			grid.data.$height = heightNew;
			
			fd.plot();
		}
	}
	
	/**
		Changes one or more properties of an edge immediately.
		Converts label strings to arrays.
		@param sourceID - the ID of the node where the edge starts
		@param targetID - the ID of the node where the edge ends
		@param properties - a data object that may contain any field from
					  data as well as "id". Data fields must start with a $.
	*/
	this.setEdgeData = function(sourceID, targetID, properties){
		
		/*
		// check arguments
		if(!validArg("setEdgeData()", sourceID, "string") ||
		   !validArg("setEdgeData()", targetID, "string")){
			return;
		}
		*/
		
		var edge = fd.graph.getAdjacence(sourceID, targetID);
		setObjectData(edge, properties);
	}
	
	/**
	Changes one or more properties of a node immediately.
	Converts displayable strings to arrays and recalculates node size.
	@param object - id of a node that needs changing
	@param properties - a data object that may contain any field from
				  node.data as well as "name" and "id". Data fields
				  must start with a $.
	 */
	this.setNodeData = function(nodeID, properties){
		
		/*
		// check argument
		if(!validArg("setNodeData()", nodeID, "string")){
			return;
		}
		*/
		
		var node = fd.graph.getNode(nodeID);
		setObjectData(node, properties);
	}
		
	/**
	Changes one or more properties of a node or edge immediately.
	Converts displayable strings to arrays and recalculates node size.
	@param object - a node or an edge that needs changing
	@param properties - a data object that may contain any field from
				  node.data as well as "name" and "id". Data fields
				  must start with a $.
	 */
	function setObjectData(object, properties){
		
		var isEdge = !!object.nodeFrom;
		var value;
		
		// iterate through properties object
		for(var field in properties){
			value = properties[field];
			
			// convert string to array if it is a displayable name
			if((field == "name" || field == "$label") && typeof value == "string"){
				value = value.split("\n");
			}
			
			// make changes to data
			if(field[0] == '$'){
				
				// treat a new font name by appending the font size as well
				if(field == "$font"){
					if(!value){
						value = "Lucida Console";
					}
					value = "px " + value;
				}
				else if(isEdge && field == '$type'){
					// apply new value to data
					object.Edge[field] = value;
				}
				
				// apply new value to data
				object.data[field] = value;
			}
			
			// make changes to name or id
			else{
				object[field] = value;
			}
		}
		
		// recalculate node dimensions in case we changed the font or name
		if(!isEdge && !object.data.$custom){
			updateNodeDimensions(object);
		}
		
		// apply visual changes
		fd.plot();
	}
	
	/**
		Recalculates the width and height of a nodeFamily and repositions the ports.
		@param node - the nodeFamily of the graph
	*/
	function updateNodeDimensions(node){
		
		var data = node.data;
		
		if(data.$type != 'nodeFamily'){
			return;
		}
		
		// reserve icon space
		var iconSpace = 0;
		if(data.$icon){iconSpace = 4;}
		// calculate width of name
		var width = 0;
		
		var font = data.$font;
		if(font){
			font = (data.$dim * 0.83) + font;
		}else{
			font = (data.$dim * 0.83) + "px Lucida Console";
		}
		
		// calculate width of longest text line
		var textArray = node.name;
		for(var t in textArray){
			width = Math.max(width, graph.getTextWidth(textArray[t], font));
		}
		
		// calculate nodeFamily dimensions
		width += (4 + iconSpace) * vertexLabelSize;
		
		// check if the width changed at all
		if(data.$width == width){
			return;
		}
		
		// calculate how much bigger/smaller the width got
		var delta = (width - node.data.$width) / 2;
		
		// update width
		node.data.$width = width;
		width /= 2;
		
		// update port positions
		var portCounter = {
			'inputPort' : 0,
			'outputPort' : 0,
			'repositoryPort' : 0};
			
		var children = data.$children,
			parentX = node.pos.getc(true).x,
			childPos;
		
		for(var c in children){
			c = children[c];
			portCounter[c.data.$type]++;
			
			childPos = c.pos.getc(true);
			
			if( parentX - childPos.x > 0){
				c.pos.setc(childPos.x - delta, childPos.y);
			}
			else{
				c.pos.setc(childPos.x + delta, childPos.y);
			}
		}
		
		// update height
		var maxPorts = Math.max(portCounter.inputPort,
								portCounter.outputPort,
								portCounter.repositoryPort);
		var size = vertexLabelSize * 2;
		
		var height = size + Math.max( 2 * size, vertexLabelSize + maxPorts * size);
		
		if(textArray){
			height = Math.max(height, (textArray.length + 2) * size);
		}
		
		// update crossBox position
		delta = (height - node.data.$height) / 2;
		if(!data.$custom){
			c = children[0];
			childPos = c.pos.getc(true);
			c.pos.setc(childPos.x, childPos.y - delta);
		}
		
		node.data.$height = height;
	}
	
	
	/**
		Extracts certain node and edge information and fuses it to two
		Strings whose elements look like this:
		NodeString: isCustomNode, width, height, parentNodeIndex
		EdgeString: srcNodeIndex, srcPortIndex, tarNodeIndex0, tarNodeIndex2, ...
		
		The resulting strings will be arguments of a listener event and can be used
		by the KIELER autolayouting algorithm.
	*/
	this.autoLayout = function(){
		
		// link port IDs to their family index
		var c, id, children, portMap = {};
		
		// list all parents of custom nodes
		var data, parentMap = {};
		iterateAllNodes(function(node){
			
			id = node.id;
			data = node.data;
			
			if(data.$custom){
				children = data.$children;
				for(c in children){
					c = children[c];
					parentMap[c.id] = id;
				}
			}
		});
		
		var nodes = [vertexLabelSize];
		var edge, edgeMap = {};
		var	nf, parentID, srcId;
		var nodeY, height;
		// map nodeIDs to indices
		iterateAllNodes(function(node){
			
			data = node.data;
			height = data.$height;
			c = data.$custom;
			nf = data.$type == 'nodeFamily';
			
			if(nf || c){
				id = node.id;
				
				// check for parent
				if(!(parentID = parentMap[id])){
					parentID = -1;
				}
				localArr = [id,
				            parentID,
							data.$width,
							height];
				
				// is Custom Node
				if(c){
					localArr.push('c'); 	// custom tag
					
					// create entry in edges
					if(!(edge = edgeMap[id])){
						edgeMap[id] = [id, -1];
					}else{
						edge[0] = id;
						edge[1] = -1;
					}
					
					// save edge info in edges-object
					adja = node.adjacencies;
					
					for(a in adja){
						a = adja[a].data.$direction;
						srcId = a[0];
						
						// fill in info for all edges that have this node
						// as a target
						if(a[1] == id){
							
							if(!edgeMap[srcId]){
								edgeMap[srcId] = ["", -1]
							}
							// add target info to edge
							edgeMap[srcId].push(id, -1);
						}
					}
				}
				
				// is NodeFamily
				else{ 
					localArr.push('f'); 	// family tag
					// inputCount, inputYPos, repoCount...
					localArr.push(0,0, 0,0, 0,0);
					
					// iterate ports and children
					children = data.$children;
					nodeY = node.pos.y;
					pIndex = -1;
					
					for(c in children){
						child = children[c];
						
						// increment node count and get relative y position of the topmost ports
						switch(child.data.$type){
							case "inputPort" : 
								if(localArr[5]++ == 0){ 
									localArr[6] = Math.round(child.pos.y - nodeY + height);
								}
								break;
						
							case "repositoryPort" : 
								if(localArr[7]++ == 0){ 
									localArr[8] = Math.round(child.pos.y - nodeY + height);
								}
								break;
								
							case "outputPort" : 
								if(localArr[9]++ == 0){ 
									localArr[10] = Math.round(child.pos.y - nodeY + height);
								}
								break;
								
							default : break;
						}
						
						if(pIndex != -1){
							// create entry in edges
							if(!(edge = edgeMap[child.id])){
								edgeMap[child.id] = [id, pIndex];
							}else{
								edge[0] = id;
								edge[1] = pIndex;
							}
							
							// save edge info in edges-object
							adja = child.adjacencies;
							
							for(a in adja){
								a = adja[a].data.$direction;
								srcId = a[0];
								
								// fill in info for all edges that have this port
								// as a target
								if(a[1] == child.id){
									
									if(!edgeMap[srcId]){
										edgeMap[srcId] = ["", -1]
									}
									// add target info to edge
									edgeMap[srcId].push(id, pIndex);
								}
							}
						}
						pIndex++;
					}
				}
				// convert node data to string and push it to Array
				nodes.push(localArr.join(" "));
			}
		});
		
		// merge arrays to strings
		var edges = [];
		var t, source;
		for(c in edgeMap){
			c = edgeMap[c];
			
			if(c.length > 2){
				source = c[0] + " " + c[1];
				t = 2;
				while(t < c.length){
					target = " " + c[t++] + " " + c[t++];
					edges.push(source + target);
				}
			}
		}
		
		/*
		// map nodeIDs to indices
		iterateAllNodes(function(node){
			
			c = node.data.$custom;
			if(c || node.data.$type == 'nodeFamily'){
				indexMap[node.id] = [index, c, node, -1];
				index++;
			}
		});
		
		// store children/parent relationships
		var children, child, pIndex;
		
		iterateAllNodes(function(node){
			children = node.data.$children;
			pIndex = 0;
			index = indexMap[node.id];
			
			for(c in children){
				// push index of child node to local Array
				c = children[c];
				child = indexMap[c.id];
				
				// child is a custom node or nodeFamily
				if(child){
					child[3] = index[0];
				}
				// child is a port
				else if(c.data.$type != "crossBox"){
					portMap[c.id] = [index[0], pIndex];
					pIndex ++;
				}
			}
		});
		
		var edges = []; 
		var nodes = [vertexLabelSize];
		
		var node, data, localArr;
		var nodeY, width, height, nodeIndex;
		var edge, adja, a, port;
		
		// write node data to string
		for(index in indexMap){
			index = indexMap[index];
			
			nodeIndex = index[0];
			node = index[2];
			data = node.data;
			nodeY = node.pos.y;
			width = Math.round(data.$width);
			height = Math.round(data.$height);
			
			// start node data with its dimensions
			localArr = ['c'];			// [0] 'c' if it is a custom node
			localArr.push(width); 		// [1] node width
			localArr.push(height);		// [2] node height
			localArr.push(index[3]);	// [3] parent index
			
			// node is a CustomNode
			if(index[1]){
				
				// extract edges
				edge = [nodeIndex, -1];
				adja = node.adjacencies;
				
				for(a in adja){
					a = adja[a].data.$direction;
					if(a[0] == node.id){
						
						// edge is connected to a port
						port = portMap[a[1]];
						if(port){
							edge.push(port[0]);
							edge.push(port[1]);
						}	
						// edge is connected to a custom node
						else{
							edge.push(indexMap[a[1]][0]);
							edge.push(-1);
						}
					}
				}
				
				if(edge.length != 2){
					edges.push(edge.join(" "));
				}
				
			}
			// node is a NodeFamily
			else{
				localArr[0] = 'f'; 	// [0] 'f' if it is a nodeFamily
				// info about ports: their number and position
				localArr.push(0); 	// [4] input count
				localArr.push(0); 	// [5] input first y-pos
				
				localArr.push(0); 	// [6] repository count
				localArr.push(0); 	// [7] repository first y-pos
				
				localArr.push(0); 	// [8] output count
				localArr.push(0); 	// [9] output first y-pos
				
				// iterate ports and children
				children = data.$children;
				pIndex = -1;
				
				for(c in children){
					child = children[c];
					switch(child.data.$type){
						case "inputPort" : 
							if(localArr[4]++ == 0){ 
								localArr[5] = Math.round(children[c].pos.y - nodeY + height);
							}
							break;
					
						case "repositoryPort" : 
							if(localArr[6]++ == 0){ 
								localArr[7] = Math.round(children[c].pos.y - nodeY + height);
							}
							break;
							
						case "outputPort" : 
							if(localArr[8]++ == 0){ 
								localArr[9] = Math.round(children[c].pos.y - nodeY + height);
							}
							break;
							
						default : break;
					}
					
					// extract port edges
					edge = [nodeIndex, pIndex];
					adja = child.adjacencies;
					for(a in adja){
						a = adja[a].data.$direction;
						
						if(a[0] == child.id){
							// edge is connected to a port
							port = portMap[a[1]];
							if(port){
								edge.push(port[0]);
								edge.push(port[1]);
							}	
							// edge is connected to a custom node
							else{
								edge.push(indexMap[a[1]][0]);
								edge.push(-1);
							}
						}
					}
					if(edge.length != 2){
						edges.push(edge.join(" "));
					}
					
					pIndex++;
				}
			}
			
			// convert node data to string and push it to Array
			nodes.push(localArr.join(" "));
			
		}*/
		
		// fuse nodes array to a single string
		nodes = nodes.join(";");
		edges = edges.join(";");
		
		callListener("autoLayout", [nodes, edges]);
	}
	
	/**
		Assumes that the graph has not been altered since the last call of autolayout()
		or saveGraphPositions().
		Reads a string containing new x,y positions, as well as dimensons which are 
		separated by spaces and moves nodefamilies to the specified positions.
		
		@param input - the layout string
	
	*/
	this.loadPositions = function(input){
		
		// check if animation is in progress
		var isAnimated = animation.enabled;
		var usePortIndex;
		var edgeInfo, nodeInfo;
		var p = 0;
		
		{
			var inputSplit = input.split("#");
			// catch Java Errors here
			if(inputSplit[0] == "Error"){
				callListener("onError", ["Error in GraphFlowLayouter.jar : " + inputSplit[1]]);
				return;
			}
			
			// check to see if we got the layout string from
			// the autolayouter, because it looks a bit different
			usePortIndex = inputSplit[0] == "autoLayout";
			if(usePortIndex){
				p++;
			}
			
			nodeInfo = inputSplit[p++].split(";");
			// bend points are optional
			if(inputSplit[p] && inputSplit[p].length > 0){
				edgeInfo = inputSplit[p].split(";");
			}

		}
		
		// prepare loop variables
		var node, data, adja, x, y;
		
		var id, width, height;
		var xOffset = -grid.data.$width/2,
			yOffset = -grid.data.$height/2;
		
		var nodeProps;
		var n, l;
		
		var endPos = new $jit.Complex(0,0);
		var g = fd.graph;
		
		// iterate node positions
		for(n = 0, l = nodeInfo.length - 1; n < l; n++){
			
			nodeProps = nodeInfo[n].split(" ");
			p = 0;
			
			// lookup node in graph
			id = nodeProps[p++];
			node = g.getNode(id);
			
			if(!node){
				callListener("onError", ["Error in loadGraphPositions() : Node with ID '"+ id + "' does not exist!"]);
			}
			else{
				// extract positions from string array
				x = parseFloat(nodeProps[p++]) + xOffset;
				y = parseFloat(nodeProps[p++]) + yOffset;
				width = parseFloat(nodeProps[p++]);
				height = parseFloat(nodeProps[p]);
				
				// is custom node
				if(node.data.$custom){
					
					// set new position and dimensions
					if(isAnimated){
						endPos.x = x;
						endPos.y = y;
						node.setPos(endPos, "end");
						node.setData('width', width, 'end');
						node.setData('height', height, 'end');
					}else{
						node.pos.setc(x, y);
						node.data.$width = width;
						node.data.$height = height;
					}
				}
				// is nodeFamily
				else{
					moveNode(node, x, y, isAnimated);
				}	
			}
		}
		
		if(edgeInfo){
			var adja;
			var from, to;
			var bl, b, bendArray, bendPoints;
			var portIndex;
			
			for(n = 0, l = edgeInfo.length - 1; n < l; n++){
				bendArray = edgeInfo[n].split(" ");
				bl = bendArray.length;
				b = 0;
				
				// get portIDs, using the familyID and the portIndex
				if(usePortIndex){
					node = g.getNode(bendArray[b++]);
					data = node.data;
					if(data.$custom){
						from = node.id;
					}else{
						portIndex = parseInt(bendArray[b]);
						from = data.$children[++portIndex].id
					}
					b++;
					
					node = g.getNode(bendArray[b++]);
					data = node.data;
					if(data.$custom){
						to = node.id;
					}else{
						portIndex = parseInt(bendArray[b]);
						to = data.$children[++portIndex].id
					}
					b++;
				
				// simply read the exact portIDs
				}else{
					from = bendArray[b++];
					to = bendArray[b++];
				}
				
				adja = g.getAdjacence(from, to);
				if(!adja){
					callListener("onError", ["Error in loadGraphPositions() : Edge between '"
					                         + from + "' and '" + to + "' does not exist!"]);
				}else{
					// new bend points exist
					if(b < bl){
						
						// read from bend point string
						bendPoints = [];
						while(b < bl){
							x = parseFloat(bendArray[b++]) + xOffset;
							y = parseFloat(bendArray[b++]) + yOffset;
							bendPoints.push({'x':x, 'y':y});
						}
						adja.data.$bendPoints = bendPoints;	
					}
					// remove bend points
					else{
						adja.data.$bendPoints = null;	
					}
				}
			}
			
		}
		
		// scale to fit the new graph
		scaleToFit(true);
		
		// animate if necessary
		if(isAnimated){
		  	
		  	animation.busyCanvas = true;
			fd.animate({  
				modes: ['node-property:width:height', 'linear', 'canvas:zoom:translate'],
				transition: $jit.Trans.Cubic.easeInOut,
				duration: animation.duration,
				onComplete: function(){animation.busyCanvas = false; fd.plot();}
			});
		}
		
		return;
	}
	
	
	/**
	Concatenates node positions and dimensions as well as
	bendPoints to a single Layout-String, similar to that
	of the GraphFlowLayouter.
	@return - the resulting string
	*/
	 this.savePositions = function(){
		var nodeLayout = [],
			edgeLayout = [];
		var nodeProps, edgeProps;
		 
		var data, nodePos;
		var a, adja;
		var b, bendPoints;
		
		// map nodeIDs to indices
		iterateAllNodes(function(node){
			data = node.data;
			
			// retrieve position and dimension of all relevant nodes
			if(node.data.$custom || data.$type == 'nodeFamily'){
				nodePos = node.pos.getc(true);
				
				nodeProps = [
				   node.id,
				   Math.round(nodePos.x),
				   Math.round(nodePos.y),
				   data.$width,
				   data.$height];
				nodeLayout.push(nodeProps.join(" "));
			}
			
			// retrieve bendPoints of edges
			adja = node.adjacencies;
			
			for(a in adja){
				data = adja[a].data;
				
				if(data.$direction[0] == node.id){
					edgeProps = [];
					
					// save connected ids
					edgeProps.push(node.id);
					edgeProps.push(data.$direction[1]);
					
					// extract bendPoints
					bendPoints = data.$bendPoints;
					
					for(b in bendPoints){
						edgeProps.push(bendPoints[b].x);
						edgeProps.push(bendPoints[b].y);
					}
					
					if(edgeProps.length > 2){
						edgeLayout.push(edgeProps.join(" "));
					}
				}
			}
		});
		
		// fuse all information to a single string
		nodeLayout = nodeLayout.join(";");
		edgeLayout = edgeLayout.join(";");
		nodeLayout += ";#" + edgeLayout + ";";
		
		callListener("onSavePositions", [nodeLayout]);
		
		return nodeLayout;
	 }
	 
	 
	
	/**
		Checks an input color-string for validation.
		Sends an onError-event if the string is not valid.
		@param callFunction - the function where the argument needs to be checked
		@param color - the color argument which is checked
		@returns - false if the string is not a well formed color string 
	*/
	this.validArgColor = function(callFunction, color){
		
		var valid = false;
		// is it a string at all ?
		if(typeof color == "string"){
			// does the string match "#RRGGBB" ?
			var regEx = /^#[a-f\d]{6}$/i;
			valid = color.match(regEx);
		}
		
		if(!valid){
			callListener("onError", ["Error in function "+ callFunction 
										+ ": Invalid color '"+ color 
										+ "'. Must be structured like '#RRGGBB'."]);
		}
		return valid;
	}
	
	
	/**
		Checks an input object for validation.
		Sends an onError-event if the Object does not have an id
		@param callFunction - the function where the argument needs to be checked
		@param node - the object argument which is checked
		@returns - false if the object does not have an id
	*/
	this.validArgNode = function(callFunction, node){
		var valid = false;
		var id = node.id; 
		
		// is it an object at all ?
		
		if(node && typeof node == "object"){
			// does the object have an id
			valid = (typeof id == "string");
		}
		
		if(!valid){
			callListener("onError", ["Error in function "+ callFunction 
										+ ": Invalid Node '"+ node 
										+ "'. Must have a field 'id'."]);
			return valid;
		}
		
		valid = id[0] != "#";
		if(!valid){
			callListener("onError", ["Error in function "+ callFunction 
										+ ": Invalid NodeID '"+ id 
										+ "'. IDs may not start with #."]);
			return valid;
		}
		
		return valid;
	}
	
	
	/**
		Checks an argument for type correctness.
		Sends an onError-event if the argument is not valid.
		@param callFunction - the function where the argument needs to be checked
		@param arg - the argument which is checked
		@param type - the type of which the argument must be
		@returns - false if the string is not a well formed color string 
	*/
	this.validArg = function(callFunction, arg, type){
		var valid = (typeof arg == type);
		if(!valid){
			callListener("onError", ["Error in function "+ callFunction 
										+ ": Invalid argument type for argument '"
										+ arg + "'. Must be of type " + type + "."]);
		}
		return valid;
	}
	
	
	/**
	Checks if a node ID is correct and non-existent in the graph.
	Sends an onError-event if the argument ID not valid.
	@param callFunction - the function where the argument needs to be checked
	@param id - the id which is checked
	@returns - false if the id is not a string or exists already
	 */
	this.validArgID = function(callFunction, id){
		
		// check if id is a string first
		if (!validArg(callFunction, id, "string")){
			return false;
		}
		
		// check if node id is occupied
		var nodeExists = fd.graph.getNode(id);
		if(nodeExists){
			callListener("onError", ["Error in function "+ callFunction 
										+ ": Duplicate Node ID '"
										+ id + "'."]);
		}
		return !nodeExists;
	}
	
	/**
		Changes the color of the grid.
		@param newColor - the new color of the grid
	*/
	this.setGridColor = function(newColor){
		/*
		if(!validArgColor('setGridColor()', newColor)){
			return;
		}
		*/
			
		grid.data.$color = newColor;
		fd.plot();
	}
	
	/**
		Changes the size of the grid.
		@param newSize - the new size of the grid
	*/
	this.setGridSize = function(newSize){
		/*
		if(!validArg('setGridSize()', newSize, 'number')){
			return;
		}
		*/
		grid.data.$dim = newSize;
		fd.plot();
		
	}
	
	/**
		Changes whether Nodes should snap to grid.
		@param snap - must be true if Nodes should snap
	*/
	this.setGridSnap = function(snap){
		grid.data.$snap = snap;
	}
	
	/**
		Changes the visibility of the grid.
		@param visibility - must be true to display the graph
	*/
	this.setGridVisible = function(visibility){
		
		isAnimated = animation.enabled && !animation.busyCanvas;
		var setter = 'current';
		
		if(isAnimated){
			setter = 'end';
		}
		
		if(visibility){
			grid.setData('alpha', 1, setter);
		}else{
			grid.setData('alpha', 0, setter);
		}
		
		if(isAnimated){
			fd.animate({  
				modes: ['node-property:alpha'],
				transition: $jit.Trans.Cubic.easeInOut,
				duration: animation.duration / 2,
				onComplete: function(){fd.plot();}
			});
		}else if(!animation.busyCanvas){
			fd.plot();
		}
	}
	
	
	/**
	 *  Returns the bounds and the exact center of the graph as an 
	 *  {x, y, left, right, top, bottom}-Object
	 *  @param checkEndPos - if true, checks the endPos instead of pos
	 *  					 of the node. This is used by autoLayout.
	 */
	this.getGraphBounds = function(checkEndPos){
		
		// init loop variables
		var data, 
			width  = 0,
			height = 0;
		
		// the position of the node which is checked
		var pos = 'pos';
		if(checkEndPos){
			pos = 'endPos';
		}
		
		var x, y;
		var b, bendPoints;
		
		// init the outer bounds of the displayed graph
		var left  	= Number.POSITIVE_INFINITY,
			right 	= Number.NEGATIVE_INFINITY, 
			top		= Number.POSITIVE_INFINITY, 
			bottom	= Number.NEGATIVE_INFINITY;
		
		// this function checks the position of bendPoints
		var checkBendPoints = function(node){
			node.eachAdjacency(function(adja){
				
				data = adja.data;
				
				// is the first node the source node?
				if(adja.nodeFrom.id == data.$direction[0]){
					bendPoints = data.$bendPoints;
					for(b in bendPoints){
						b = bendPoints[b];
						
						x = b.x;
						y = b.y;
						
						left = Math.min(left, x);
						right = Math.max(right, x);
						top = Math.min(top, y);
						bottom = Math.max(bottom, y);
					}
					
				}
			});
		};
		
		// check nodeFamilies for position
		iterateAllNodes(function(node){
			data = node.data;
			
			if(data.$custom || data.$type == "nodeFamily"){
			
				x = node[pos].x;
				y = node[pos].y;
				
				width  = data.$width/2;
				height = data.$height/2;
				
				left = Math.min(left, x - width);
				right = Math.max(right, x + width);
				top = Math.min(top, y - height);
				bottom = Math.max(bottom, y + height);
				
				if(data.$custom){
					checkBendPoints(node);
				}
			}
			// is a port
			else{
				checkBendPoints(node);
			}
		});
		
		var midX = (left + right) / 2,
			midY = (top + bottom) / 2;
		
		var bounds = 
			{
				'x' : midX,
				'y' : midY,
				'left' : left,
				'right' : right,
				'top' : top,
				'bottom' : bottom
			};
		
		return bounds;
	}
	
	
	/**
	 *  Returns the exact center of the canvas where the screen is currently located
	 *  as an {x, y}-Object
	 */
	this.getScreenCenter = function(){
		var canvas = fd.canvas,
			scale = canvas.scaleOffsetX,
			centerX = -canvas.translateOffsetX / scale,
			centerY = -canvas.translateOffsetX / scale;
		
		return {'x' : centerX, 'y' : centerY};
	}
	
	/**
		Scales and centers the graph to fit the space provided by the container
		in which it is drawn.
		@param forLoadPositions - if exists and true, the animation won't be called,
				   and the endPos of nodes will be checked. These are
				   the positions where the nodes will be after the 
				   animation.
	*/
	this.scaleToFit = function(forLoadPositions){
		// do nothing if the graph is empty
		if(fd.graph.nodes.length == 2){
			return;
		}
		
		var isAnimated = animation.enabled && !animation.busyCanvas;
		
		// get graph bounds
		var bounds = getGraphBounds(forLoadPositions && isAnimated);
		
		// scaling values
		var canvas = fd.canvas,
			oldScale = canvas.scaleOffsetX,
			scaleX = grid.data.$width / (bounds.right - bounds.left + 4*vertexLabelSize),
			scaleY = grid.data.$height / (bounds.bottom - bounds.top + 4*vertexLabelSize)
			scale = Math.min(scaleX, scaleY);
		
		// translation values
		var midX = - bounds.x,
			midY = - bounds.y;
		
		
		if(isAnimated){
			
			// set animation parameters
			fd.canvas.translateOffsetXEnd = midX;
		  	fd.canvas.translateOffsetYEnd = midY;
		  	
		  	canvas.scaleOffsetOld = oldScale;
		  	canvas.scaleOffsetEnd = scale;
		  	
		  	if(!forLoadPositions){
		  		animation.busyCanvas = true;
				fd.animate({  
					modes: ['canvas:zoom:translate'],
					transition: $jit.Trans.Cubic.easeInOut,
					duration: animation.duration,
					onComplete: function(){animation.busyCanvas = false; fd.plot();}
				});
		  	}
			
		}else{
			// translate to the center of the graph and scale
			midX -= fd.canvas.translateOffsetX / oldScale;
			midY -= fd.canvas.translateOffsetY / oldScale;
			scale /= oldScale;
			
			fd.canvas.translate(midX, midY);
			canvas.scale(scale, scale);
			
			if(!animation.busyCanvas){
				fd.plot();
			}
		}
		
		// show text only if it is big enough
		canvas.showLabels = (canvas.scaleOffsetX > canvas.labelThreshold);
	}
	
	
	
	/**
	Assigns an iconPath to a type of node.
	@param nodeType - the type of the node. Typically 'Repository',
					  'Filter', or 'Reader'
	@param path - the path to the image. If null, an existing
				  icon will be deleted
	@param resizeImmediately - if true, this function will not wait
				  until the image is loaded and instead resize all
				  affected nodes, regardless of if the image is loaded
				  successfully or not
	*/
	this.setNodeIcon = function(nodeType, path, resizeImmediately){
		// TODO: give refrence instead of icon to nodes
		/*
		if(!validArg('setNodeIcon()', nodeType, 'string')){
			return;
		}
		*/
	
		if(path == null){
			delete nodeIcons[nodeType];
			
			// remove icons from displayed graph and shrink nodes
			var data;
			
			iterateAllNodes(function(node){
				// get data
				data = node.data;
				
				// update node width and port positions
				if(data.$nodeType == nodeType){
					delete data.$icon;
					updateNodeDimensions(node);
				}
			});
			
			return;
		}
		
		/*
		if(!validArg('setNodeIcon()', path, 'string')){
			return;
		}
		*/
		
		var prevImage = nodeIcons[nodeType];
		
		var icon;
		icon = new Image();
		
		// load error
		icon.onerror = function(){
			callListener("onError", ["Error in function setNodeIcon(): Could not load '"+ path +"'."]);
		};
		
		// load success
		icon.onload = function(){
			// this may happen on IE, when linking to images via web
			if(!icon.complete){
				callListener("onError", ["Error in function setNodeIcon(): Timeout while loading '"+ path +"'. Please try again!"]);
				return;
			}
			nodeIcons[nodeType] = icon;
			resizeNodes(icon, !resizeImmediately);
			fd.plot();
		};
		
		
		var resizeNodes = function(newIcon, resize){
			// if nodes were already created, we need to
			// adjust their width and iconPath
			var data;
			iterateAllNodes(function(node){
				// get data
				data = node.data;
				// update node width and port positions
				if(data.$nodeType == nodeType){
					data.$icon = newIcon;
					if(resize && !prevImage){
						updateNodeDimensions(node);
					}
				}
			});
		};
		
		if(resizeImmediately){
			nodeIcons[nodeType] = "emptyImage";
			resizeNodes("emptyImage", true);
		}
		
		// attempt to load the image
		icon.src = path;
	}
	
	/**
	Changes the appearance of the mouse cursor. 
	@param newCursor - 	a String-name of the new cursor. 
						If null, restore the cursor.
	*/
	this.setMouseCursor = function(newCursor){
		if(newCursor == null){
			fd.canvas.getElement().style.cursor = '';
		}
		/*
		else if(!validArg('setMouseCursor()', newCursor, 'string')){
			return;
		}
		*/
		else{
			fd.canvas.getElement().style.cursor = newCursor;
		}
	}
	
	
	/**
		Adds a functionality to the following MouseEvents:
			onClick,
			onRightClick,
			onMouseMove,
			onMouseEnter,
			onMouseLeave,
			onDragStart,
			onDragMove,
			onDragCancel,
			onDragEnd
		Or one of these events:
			onError(errorMessage)
			onCreateEdge(sourceNode, targetNode, sourcePort, targetPort) : boolean,
			onCreateNode(node) : boolean,
			onRemoveEdge(sourceNode, targetNode, sourcePort, targetPort) : boolean,
			onRemoveNode(node) : boolean
		@param eventName - the name of the mouse event ( see list above)
		@param listenerFunction - a function that may alter the graph
	*/
	this.addListener = function(eventName, listenerFunction){
		/*
		// check arguments
		if(!validArg('addListener()', eventName, 'string') ||
			!validArg('addListener()', listenerFunction, 'function')){
				return;
		}
		*/
	
		if(listener[eventName] == undefined){
			listener[eventName] = [];
		}
		listener[eventName].push(listenerFunction);
	}
	
	/**
		Calls all listeners that are registered under the eventName.
		@param eventName - see 'addListener()'
		@param arguments - an array of arguments. For mouseEvents it is [node, eventInfo, e],
						   for nodeEvents it is [node], 
						   and for edgeEvents it is [sourceNode, targetNode, sourcePort, targetPort]
		@return true if an editor action is permitted
	*/
	function callListener(eventName, arguments){
		var lArr = listener[eventName];
		
		if(!lArr){
			return true;
		}
		var permitted = true;
		var testval;
		
		// iterate through all listener this event concerns
		for(var e in lArr){
			testval = lArr[e].apply(lArr[e], arguments);
			
			if(testval == false){
				permitted = false;
			}
		}
		return permitted;
	}
	
	/**
	Adds a listener that prohibits edge creation if...
		... an edge with the same source and target exists.
		...	the target node is the source node.
		... the source and target node are of the same port type.
		... a repository node is not connected to a repository and
			vice versa.
	*/
	this.addEdgeConstraints = function(){
		addListener("onCreateEdge", 
		function(sourceFamily, targetFamily, sourcePort, targetPort){
			// remove Edge if it does not lead to an inputPort
			if(targetPort.data.$type != "inputPort"){
				return false;
			}
		
			// remove Edge if the connected nodes share the same type
			if(sourcePort.data.$type == targetPort.data.$type){
				return false;
			}
			
			var goesToRepo = (targetFamily.data.$nodeType == "Repository"),
				isRepoPort = (sourcePort.data.$type == "repositoryPort");
			
			// remove Edge if it is falsely connected to repository ports
			if(goesToRepo != isRepoPort){
				return false;
			}
			
			// remove duplicate Edge
			var adja = sourcePort.adjacencies,
				duplicate = false;
			for(var a=0; a < adja.length; a++){
				if(adja[a].nodeTo == targetPort.id){
					if(duplicate){
						return false;
					}
					else{
						duplicate = true;
					}
				}
			}
			return true;
		});
	}
	
	/**
	Iterates through all nodes and ports, calling a given function for each one.
	@param nodeFunction - a single parameter function, called on a node.
	*/
	this.iterateAllNodes = function(nodeFunction){
		/*
		// check arguments
		if(!validArg('iterateAllNodes()', nodeFunction, 'function')){
			return;
		}
		*/
		
		// filter out dummy nodes
		fd.graph.eachNode(function(node){
			if(node && node.id[0] != "#"){
				nodeFunction(node);
			}
		});
		
	}
	
	/**
	Changes the colors of the entire graph.
	@param nodeType - the name of the nodeType, whose fill and stroke color are changed
	@param fillColor - the fill color of nodes. If null, restore Node Color from global Array.
	@param strokeColor - the stroke color of nodes. If null, restore Stroke Color from global Array.
	@param edgeColor - the color of edges. If null, restore Edge Color from global Array.
	@param highlightColor - the of highlighted objects. If null, restore Edge Color from global Array.
	*/
	this.setNodeStyle = function(nodeType, fillColor, strokeColor, edgeColor, focusColor){
		if(!validArg('setNodeStyle()', nodeType, 'string')){
			return;
		}
		
		if(fillColor /*&& validArgColor('setNodeStyle()', fillColor)*/){
			nodeFillColor[nodeType] = fillColor;
		}
		if(strokeColor /*&& validArgColor('setNodeStyle()', strokeColor)*/){
			nodeStrokeColor[nodeType] = strokeColor;
		}
		if(edgeColor /*&& validArgColor('setNodeStyle()', edgeColor)*/){
			edgeColor = edgeColor;
		}
		if(focusColor /*&& validArgColor('setNodeStyle()', focusColor)*/){
			edgeColorFocus = focusColor;
			nodeColorFocus = focusColor;
		}
		
		var data, a, al, adja;
		// change colors
		iterateAllNodes(function(node){
			data = node.data;
			
			if(data.$type == nodeType && node != markedNode.node){
				data.$fillColor = nodeFillColor[nodeType];
				data.$color = nodeStrokeColor[nodeType];
			}
			// change edge colors
			if(edgeColor){
				//TODO: this
				/*
				adja = json[n].adjacencies; // TODO: data.adjacencies?
				for(a = 0, al = adja.length; a < al; a++){
					adja[a].data.$color = edgeColor;
				}
				*/
			}
		});
	}
	
	/**
	Permits or prohibits node and edge deletion/creation and hides the deletion-cross
	accordingly.
	@param deleteCreate - if true, one can change the graph in any way
	@param move - if true, one can change the graph in any way
	*/
	this.setReadOnly = function(deleteCreate, move){
				
		if(deleteCreate != readOnly.deleteCreate){
			// set crossBoxes (in)visible
			var data;
			
			iterateAllNodes(function(node){
				data = node.data;
				if(data.$type == "crossBox"){
					data.$visible = deleteCreate;
				}
			});
			
			readOnly.deleteCreate = deleteCreate;
		}
		
		if(move != undefined){
			readOnly.move = move;
		}
		fd.plot();
	}
		
	/**
	 Adds a Node to the graph.
	 @param xPosition - the horizontal position of the added node. 0 is the center of the canvas' starting position
	 @param xPosition - the vertical position of the added node. 0 is the center of the canvas' starting position
	 @param nodeFamily - describes some properties of the node: (id, name, nodeClass, tooltip)
	 @param repositoryPorts - an array of properties for the input ports (id, name, tooltip)
	 @param inputPorts - an array of properties for the input ports (id, name, tooltip)
	 @param outputPorts - an array of properties for the output ports (id, name, tooltip)
	 @param nodeType - a string-identifier for a certain group of nodefamilies eg. Filter, Repository
	 @param forced - if true, no event listener will be called and the node will be created no matter what
	 */
	this.addNode = function(xPosition, yPosition, nodeFamily, repositoryPorts, inputPorts, outputPorts, nodeType, forced){
	// determine array size
	var countRepo = 0,
		countInput = 0,
		countOutput = 0,
		iconSpace = 0;
		
	if(repositoryPorts){
		countRepo = repositoryPorts.length;
		// check if it is an array
		if(/*!validArg(caller, repositoryPorts, 'object') ||*/ !countRepo){
			countRepo = 0;
		}
	}
	if(inputPorts){
		countInput = inputPorts.length;
		// check if it is an array
		if(/*!validArg(caller, inputPorts, 'object') ||*/ !countInput){
			countInput = 0;
		}
	}
	if(outputPorts){
		countOutput = outputPorts.length;
		// check if it is an array
		if(/*!validArg(caller, outputPorts, 'object') ||*/ !countOutput){
			countOutput = 0;
		}
	}
	
	/*
	// check for other illegal arguments
	var caller = 'addNode()';
	if(!validArg(caller, xPosition, 'number')){
		xPosition = 0;
	}
	if(!validArg(caller, yPosition, 'number')){
		yPosition = 0;
	}
	if(!validArg(caller, nodeType, 'string')){
		nodeType = 'Filter';
	}
	if(!validArgNode(caller, nodeFamily)){
		return;
	}
	if(!validArgID(caller, nodeFamily.id)){
		return;
	}
	*/
	
	var maxPorts = Math.max(countInput, countRepo+countOutput+0.5),
		size = vertexLabelSize*2;
		
	// insert some space between repo ports and output ports, only if both are present
	if(maxPorts > countInput && (countRepo == 0 || countOutput == 0)){
		maxPorts-= 0.5;
	}
	
	// check if this nodeType has a defined icon
	if(nodeIcons[nodeType]){
		iconSpace = 4;
	}
	
	// set up font
	var font = nodeFamily.font;
	if(!font){ font = "Lucida Console";}
	font = "px " + font;
	
	// calculate node family dimensions
	var width = 0;	
	var height = size + Math.max( 2 * size, vertexLabelSize + maxPorts * size);
	
	// calculate width of longest text line
	if(nodeFamily.name){
		var textArray = nodeFamily.name.split("\n");
		for(var t in textArray){
			width = Math.max(width, graph.getTextWidth(textArray[t], (size * 0.83) + font));
		}
		
		height = Math.max(height, (textArray.length + 2) * vertexLabelSize * 2);
	}
	width += (4 + iconSpace) * vertexLabelSize;

	// add big node box
	// change node color, depending on its type
	var nodeColor = nodeFillColor["nodeFamily"],
		strokeColor = nodeStrokeColor["nodeFamily"];
	
	var newNode = {
		  "data": {
			"$dim": size,					
			"$type": "nodeFamily",
			"$icon": nodeIcons[nodeType],
			"$width": width,
			"$height": height,
			"$color": strokeColor,
			"$fillColor": nodeColor,
			"$nodeType": nodeType,
			"$children" : [],
			"$font": font,
			"$movable" : true,
			"$tooltip": nodeFamily.tooltip
		  }, 
		  "id": nodeFamily.id, 
		  "name": textArray//nodeFamily.name
	};
	// add closeButton
		var closeButton ={
			  "data": {
				"$dim": size,					
				"$type": "crossBox",
				"$color": strokeColor,
				"$fillColor": nodeColor,
				"$visible": readOnly.deleteCreate,
				"$movable": false
			   }, 
			  "id": nodeFamily.id+".close", 
			  "name": "x"
		};
	
	// set up some more data
	var g = fd.graph;
	g.addNode(newNode);
	g.addNode(closeButton);
	newNode = g.getNode(nodeFamily.id);
	closeButton = g.getNode(closeButton.id);
	closeButton.data.$family = newNode;
	
	// set node positions
	newNode.pos.setc(xPosition, yPosition);
	closeButton.pos.setc(xPosition + (width - vertexLabelSize) / 2 , yPosition - height / 2 + size);
	newNode.data.$children.push(closeButton);
	
	// Loop Info: 
	// This loop adds all port types to the temporary port array.
	// For fast position computation, the relative position
	// of the first port to the nodeFamily-box is computed for each port type
	// it is stored in relativeX, relativeY
	
	for(var portType = 0; portType < 3; portType++){
		var x, y, relativeX, relativeY,
			loopPorts, loopType;
		switch(portType){
			case 0: // input ports
				relativeX = -width/2;
				relativeY = (0.75 - countInput/2 )*size;
				loopPorts = inputPorts;
				loopType = "inputPort";
				break;
		
			case 1: // repository ports
				var repoSpace = 0.5;
				if(countOutput == 0){
					repoSpace = 0;
				}
				relativeX = width/2;
				relativeY = (0.75 - (countOutput+countRepo+repoSpace)/2 )*size;
				loopPorts = repositoryPorts;
				loopType = "repositoryPort";
				break;
				
			case 2: // output ports
				var repoSpace = 0.5;
				if(countRepo == 0){
					repoSpace = 0;
				}
				relativeX = width/2;
				relativeY = (0.75 - (countOutput-countRepo-repoSpace)/2 )*size;
				loopPorts = outputPorts;
				loopType = "outputPort";
				break;
		}
		
		// continue only if ports exist
		if(loopPorts != null && loopPorts.length > 0){
		
			x = xPosition + relativeX;
			y = yPosition + relativeY;
			var port, portID;
			for(var p=0; p < loopPorts.length; p++){
				port = loopPorts[p];
				
				// only attempt to add a node if it has a valid id
				portID = nodeFamily.id+"."+port.id;
				
				//if(validArgNode(caller, port) && validArgID(caller, portID)){
					var newPort ={
						  "adjacencies": [], 
						  "data": {
							"$dim": size,					
							"$type": loopType,
							"$color": nodeStrokeColor[loopType],
							"$fillColor": nodeColor,
							"$tooltip": port.tooltip,
							"$movable" : false,
							"$symbol": port.symbol}, 
						  "id": portID, 
						  "name": port.name};
					
					g.addNode(newPort);
					newPort = g.getNode(newPort.id);
					newPort.pos.setc(x, y);
					newNode.data.$children.push(newPort);
					
					y += size;
				//}
			}
		}
	}
	
		// call listener
		if(!forced){
			callListener("onCreateNode", [newNode]);
		}
		fd.plot();
		
		return newNode;
	}

	/**
		Removes a Node and all its attached edges from the graph.
		@param nodeFamily - the node which is to be removed
		@param forced - if true, the node will be removed unconditionally and no
						event listeners will be called
	*/
	this.removeNode =  function(nodeFamily, removeChildren, forced){
		
		/*
		// check for valid arguments
		if(!validArgNode("removeNode()", nodeFamily)){
			return;
		}
		*/
		
		// call listener
		if(!forced && !callListener("onRemoveNode", [nodeFamily])){
			return;
		}
		
		// unmark node
		if (markedNode && markedNode.node.id == nodeFamily.id){
			markedNode = null;
		}
		
		var g = fd.graph;
		
		// remove all children from graph
		var children = nodeFamily.data.$children;
		if(removeChildren && children){
			var child;
			for(var c in children){
				g.removeNode(children[c].id);
			}
			
		}
		// remove node itself
		g.removeNode(nodeFamily.id)
		
		fd.plot();
		return;
	 }
	 
	 
	 /**
		Moves a nodeFamily and its children to the specified position.
		@param node - the node to be moved
		@param xPos - the horizontal position of the moved node. 0 is the center of the canvas' starting position
		@param yPos - the vertical position of the moved node. 0 is the center of the canvas' starting position
		@param smoothMove - used for smooth animation in autoLayout
	 */
	 function moveNode(node, x, y, smoothMove){
	
		var data = node.data;
		 
		// snap to grid
		if(!smoothMove && grid.data.$snap && data.$movable){
			var gridDim = grid.data.$dim,
				left = (x - data.$width/2),
				top = (y - data.$height/2),
				right = (x + data.$width/2),
				bottom = (y + data.$height/2),
				
				offLeft = Math.abs(left % gridDim),
				offTop = Math.abs(top % gridDim),
				offRight = Math.abs(right % gridDim),
				offBottom = Math.abs(bottom % gridDim),
				threshold = gridDim/6;
		
			if(top > 0){
				offTop = gridDim - offTop;
			}
			if(bottom > 0){
				offBottom = gridDim - offBottom;
			}
			if(left > 0){
				offLeft = gridDim - offLeft;
			}
			if(right > 0){
				offRight = gridDim - offRight;
			}
			
			// snap to closest gridline
			if(offLeft < threshold) {
				x += offLeft;
			}
			if(offRight < threshold) {
				x += offRight;
			}
			else if(offLeft > gridDim - threshold){
				x -= gridDim-offLeft;
			}
			else if(offRight > gridDim - threshold){
				x -= gridDim-offRight;
			}
			
			if(offTop < threshold) {
				y += offTop;
			}
			
			else if(offTop > gridDim - threshold){
				y -= gridDim-offTop;
			}
			
			else if(offBottom < threshold) {
				y += offBottom;
			}
			
			else if(offBottom > gridDim - threshold){
				y -= gridDim-offBottom;
			}
		}
	 
		// parent position
		var pos = node.pos.getc(true);
		 
		var deltaX = x - pos.x,
			deltaY = y - pos.y;
		
		// move parent
		if(smoothMove){
			node.setPos(new $jit.Complex(x, y), "end");
		}else{
			node.pos.setc(x, y);
		}
		
		 
		// move children in relation to their parent
		var children = node.data.$children,
			child;
		if(children){
			 for(var c in children){
				 
				 child = children[c];
				 pos = child.pos.getc(true);
				 moveNode(child, pos.x + deltaX, pos.y + deltaY, smoothMove);
			 }
		}
	 }

  
  /**
   *  Adds a directed edge between two nodes to the graph.
   *  @param sourceID - the id of the node from which the edge starts
   *  @param targetID - the id of the node where the edge ends
   *  @param edgeLabel - the label of the edge. Can be null if no label is required
   *  @param edgeData - a data object which may overwrite the standard edge data
   *  @param forced - if true, creates the edge unconditionally and does not trigger 
					  a listener event
   *  @return true - if the edge was successfully added
   */
	this.addEdge = function(sourceID, targetID, edgeLabel, edgeData, forced){
		/*
		// check for valid arguments
		if(!validArg("addEdge()", sourceID, "string")){
			return false;
		}
		if(!validArg("addEdge()", targetID, "string")){
			return false;
		}
		*/
		if(edgeLabel && typeof edgeLabel == "string"){
			edgeLabel = edgeLabel.split('\n');
		}
		/*
		if(edgeLabel){
			if(typeof edgeLabel == 'object'){
				// do nothing
			}else if(validArg("addEdge()", edgeLabel, "string")){
				edgeLabel = edgeLabel.split('\n');
			}else{
				edgeLabel = null;
			}
		}
		*/
		
		// look up the nodes which are to be connected
		var g = fd.graph;
		var sourcePort = g.getNode(sourceID),
			targetPort = g.getNode(targetID);
		
		// check if ports were found
		if(!sourcePort){
			callListener("onError", ["Error in function addEdge(): The source node '" + sourceID + "' does not exist."]);
			return false;
		}
		if(!targetPort){
			callListener("onError", ["Error in function addEdge(): The target node '" + targetID + "' does not exist."]);
			return false;
		}
		
		// the standard data set, which may be manipulated by event listeners
		var data =
			{"$direction" : [ sourceID, targetID ],
			 "$color" : edgeColor,
			 "$type" : "flowArrow",
			 "$offsetFrom" : 
				   customEdgeMap[sourcePort.data.$type].getOffset(sourcePort, targetPort, true, edgeData),
		     "$offsetTo" :
		    	   customEdgeMap[targetPort.data.$type].getOffset(sourcePort, targetPort, false, edgeData)};
		
		// overwrite data if it was specified
		if(edgeData){
			for(var d in edgeData){
				data[d] = edgeData[d];
			}
		}
		
		// add label to data
		if(edgeLabel){
			data.$label = edgeLabel;
        }
		
		// if it is a dangling edge, no listener needs to be called
		if(sourceID == mouseNode.id || targetID == mouseNode.id){
			data.$isMouseEdge = true;
			
			// apply changes
			fd.graph.addAdjacence(sourcePort, targetPort, data);
			fd.plot();
			return fd.graph.getAdjacence(sourcePort, targetPort);
		}
		
		delete data.$isMouseEdge;
		
		// call listener if the edge is not connected to the dummy node and check for permission
		if(!forced){
			// look for nodeFamilies if connected nodes are ports
			var sourceFamily, targetFamily;
			if(!(sourcePort.data.$custom && targetPort.data.$custom)){
				var children;
				iterateAllNodes(function(node){
					
					// look for nodeFamilies
					if(node.data.$type == "nodeFamily"){
						children = node.data.$children;
						
						// look through the ports
						for(var c in children){
							if(children[c].id == sourceID){
								sourceFamily = node;
							}
							else if(children[c].id == targetID){
								targetFamily = node;
								data.$targetFamily = node.id
							}
						}
					}
				});
			}
			
			if (!callListener("onCreateEdge", [sourceFamily, targetFamily, sourcePort, targetPort, data])){
				return false;
			}
		}
		
		// apply changes
		fd.graph.addAdjacence(sourcePort, targetPort, data);
		fd.plot();
		return fd.graph.getAdjacence(sourcePort, targetPort);
	}
	
	 /**
	   * Creates a dangling edge between a node and the mouse pointer.
	   * If such an edge exists already, create an edge between both nodes
	   * that were passed to this function.
	   * 
	   * @param node - the node which is to be connected to the mouse pointer
	   * @param isIncoming - if true, the edge will point towards the node
	   * @param label - an optional edge label
	   * @param data - a data object which may overwrite the standard edge data
	   */
	  this.connectEdge = function(nodeID, isIncoming, label, data){
		
		// second node clicked?
		if(selectedNode){
			var label = selectedNode.label;
			var newEdge = false;
			
			// do we create an edge or simply discard it
			if(nodeID){
				
				// determine direction of edge
				if(selectedNode.isIncoming){
					newEdge = addEdge(nodeID, selectedNode.id, label, data);
				}
				else{
					newEdge = addEdge(selectedNode.id, nodeID, label, data);
				}
			}
			
			// remove mouse edge
			if(selectedNode.isIncoming){
				removeEdge(mouseNode.id, selectedNode.id, true);
			}else{
				removeEdge(selectedNode.id, mouseNode.id, true);
			}
			
			selectedNode = null;
			
			return newEdge;
		}
		// first node clicked
		else{
			selectedNode = {"id": nodeID, "isIncoming" : isIncoming, "label" : label};
			
			navi.isDraggingEdge = true;
			// determine direction of mouse edge
			if(isIncoming){
				return addEdge(mouseNode.id, nodeID, label, data);
			}else{
				return addEdge(nodeID, mouseNode.id, label, data);
			}
		}
	  }
  
  
  /**
		Sets or removes BendPoints of an Edge.
		@param sourceID - the ID of the connected outputPort
		@param targetID - the ID of the connected inputPort
		@param bendPoints - an array of {"x":..., "y":...} objects
  */
  this.setBendPoints = function(sourceID, targetID, bendPoints){
	  	/*
		// check for valid arguments
		if(!validArg("setBendPoints()", sourceID, "string")){
			return false;
		}
		if(!validArg("setBendPoints()", targetID, "string")){
			return false;
		}
		*/
		
		// change data
		fd.graph.getAdjacence(sourceID, targetID).data.$bendPoints = bendPoints;
		
		// plot visual changes
		fd.plot();
  }
  
  /**
	Removes the edge between two nodes.
	@param sourceID - the id of the source Node
	@param targetID - the id of the target Node
	@param forced - if true, removes the edge unconditionally and
					does not trigger a listener event
	*/
  this.removeEdge = function(sourceID, targetID, forced){
	/*
	// check for valid arguments
	if(!validArg("removeEdge()", sourceID, "string")){
		return;
	}
	if(!validArg("removeEdge()", targetID, "string")){
		return;
	}
	*/
  
	// look up the source node and the nodeFamilies of both nodes
	var g = fd.graph,
		source = g.getNode(sourceID), 
		target = g.getNode(targetID);
	
	var adja = g.getAdjacence(sourceID, targetID);
	
	// nothing to delete?
	if(!source || !adja){
		return;
	}
		
	if(!forced && sourceID != mouseNode.id && targetID != mouseNode.id){
		
		// look for nodeFamilies if connected nodes are ports
		var targetFamily = g.getNode(adja.data.$targetFamily);
		var sourceFamily;
		if(!(source.data.$custom && target.data.$custom)){
			var children;
			iterateAllNodes(function(node){
				
				// look for nodeFamilies
				if(node.data.$type == "nodeFamily"){
					children = node.data.$children;
					
					// look through the ports
					for(var c in children){
						if(children[c].id == sourceID){
							sourceFamily = node;
							break;
						}
					}
				}
			});
		}
		// call listener and ask for permission to delete
		if(!callListener("onRemoveEdge", [sourceFamily, targetFamily, source, target])){
			return;
		}
	}
	
	// apply changes
	fd.graph.removeAdjacence(sourceID, targetID);
	fd.plot();
  }
  
  /**
  Returns a node from the graph
  @param nodeID - the id of the desired node
  @return - the node object or null if the node does not exist
  */
  this.getNode = function(nodeID){
	/*
	// check for valid arguments
	if(!validArg("getNode()", nodeID, "string")){
		return;
	}
	*/
	
	return fd.graph.getNode(nodeID);
  }
  
  
  
  /**
	Highlights exactly one node or edge, and un-highlights the previously highlighted
	node or edge.
	@param node - the node or edge object which will be highlighted
	@param noAnim - does not animate the highlight. Used by loadPositions()
	*/
  function setHighlight(node, noAnim){
	  
	  if(animation.busyCanvas){
		  return;
	  }
	  
	  var setter = 'current';
	  var isAnimated = animation.enabled;
	  if(isAnimated){
		  setter = 'end';
	  }
	  
		// animation parameters
		var trans = $jit.Trans.Quint.easeOut, 
			dur = animation.duration / 4;
		
		// clean up the old highlight
		if(hover){
			
			if(hover.nodeFrom){
				hover.setData('lineWidth', 1, setter);
				hover.setData('color', edgeColor, setter);
			}
			else{
				var type = hover.data.$type;
				if(markedNode != null && hover.id == markedNode.node.id){
					hover.setData('color', markedNode.strokeColor, setter);
				}
				else{
					hover.setData('color', nodeStrokeColor[type], setter);
				}
				// decrease size of ports and cross button
				if(type == 'inputPort' || type == 'outputPort' || type == 'repositoryPort'){
					hover.setData('dim', vertexLabelSize*2, setter);
				}
				
			}
			hover = null;
		}
		
		// do we highlight something new?
		if(node != null){
			if(node.nodeFrom != undefined){
				node.setData('color', edgeColorFocus, setter);
				node.setData('lineWidth', 2, setter);
			}else{
				node.setData('color', nodeColorFocus, setter);
				
				// increase size of ports and cross button
				var type = node.data.$type;
				if(type == 'inputPort' || type == 'outputPort' || type == 'repositoryPort'){
					node.setData('dim', vertexLabelSize*3, setter);
				}
			}
			hover = node;
		}else{
			// slower fade away transition
			trans = $jit.Trans.Quint.easeIn;
			dur *= 2;
		}
		
		if(noAnim){
			return;
		}
		
		// Animations overwrite each other. If we let this animation
		// execute while we delete an edge, the onComplete() function
		// of the edge animation will not be called
		if(isAnimated){
			fd.animate({  
					modes: ['edge-property:lineWidth:color',
							'node-property:color:dim'],
					transition: trans,  
					duration: dur,
					onComplete: function(){fd.plot();}
			});
		}else{
			fd.plot();
		}
		
  }
  
  /**
	Highlights exactly one node, and un-highlights the previously highlighted
	node.
	@param node - the node or edge object which will be highlighted.
				  Can be null if no new node shall be highlighted
	@param strokeColor - the color of the node stroke or a port's fillcolor.
						 Can be null for no change.
	@param fillColor - the fill color of a nodefamily
						Can be null for no change.
	*/
  this.markNode = function(node, strokeColor, fillColor){
  
	  	var isAnimated = animation.enabled && !animation.busyCanvas;
	  	var setter = 'current';
	  	if(isAnimated){
	  		setter = 'end';
	  	}
	  	
	  	// do not mark edges
	  	if(node.nodeFrom){
			return;
		}
	  	
		// check for valid arguments
		/*
		if(node && !validArgNode("markNode()", node)){
			return;
		}
		if(strokeColor && !validArgColor("markNode()", strokeColor)){
			return;
		}
		if(fillColor && !validArgColor("markNode()", fillColor)){
			return;
		}
		*/
		
		// clean up the old marking
		if(markedNode){
			
			if(markedNode.node.nodeFrom){
				// ignore edges
			}
			else{
				var type = markedNode.node.data.$type;
				markedNode.node.setData('color', nodeStrokeColor[type], setter);
				
				// change FillColor only if it exists
				if(nodeFillColor[type]){
					markedNode.node.setData('fillColor', nodeFillColor[type], setter);
				}
			}
			markedNode = null;
		}
		
		// do we highlight something new?
		if(node){
		
			if(node.nodeFrom != undefined){
				// ignore edges
			}else{
				if(strokeColor != null){
					node.setData('color', strokeColor, setter);
				}
				
				// change FillColor only if it exists
				if(fillColor != null  && nodeFillColor[node.data.$type]){
					node.setData('fillColor', fillColor, setter);
				}
			}
			markedNode = { 'strokeColor' : strokeColor, 'fillColor' : fillColor, 'node' : node};
		}
		
		// animate
		if(isAnimated){
			fd.animate({  
				modes: ['edge-property:color',
						'node-property:color'],
				transition: $jit.Trans.Quint.easeOut,  
				duration: animation.duration / 2,
				onComplete: function(){fd.plot();}
			}); 
		}else if(!animation.busyCanvas){
			fd.plot();
		}
  }
  
  /**
   * @returns true if a mouse edge exists
   */
  this.isEdgeDragged = function(){
	  return navi.isDraggingEdge;
  }
  
  /**
   * Tells the Graph information about a CustomNode, as it may be implemented using Plugins.
   * @param nodeType - the name of the nodeType
   * @param fillColor - 
   * @param strokeColor - color of the outline of the custom node
   * @param edgeOffsetFun(ajacency, isSource) - a function which calculates the offset of a connecting edge
   * @param edgeType - the type of edge which is generated when calling connectEdge() on a custom node
   */
  this.registerCustomNode = function(nodeType, fillColor, strokeColor, edgeOffsetFun){
	  
	    // define the color
		setNodeStyle(nodeType, fillColor, strokeColor);
		
		// define the edge and offset
		customEdgeMap[nodeType] = {'getOffset' : edgeOffsetFun};
	}
  
    
  /**
	Initializes the Graph, enabling canvas navigation and tooltips.
	Also the graph is initialized with a dummy mouse node and
	what happens on various mouse events is determined.
	*/
  this.initGraph = function(jsonArray){
	var json = [];
	  
	// init json Array
	if(jsonArray){
		json = jsonArray;
	}
	
	// define mouseNode
	json.push(
		{
		  "adjacencies": [], 
		  "data": {
			"$type": "none"
		  }, 
		  "id": "#DUMMY_MOUSE_NODE"
		},
		
		// define gridNode
		{ 
		'data': {
			'$dim' : vertexLabelSize*4,
			'$type'	 : 'grid',
			'$color'  : '#AAAAFF',
			'$width' : document.getElementById(visContainer).clientWidth,
			'$height': document.getElementById(visContainer).clientHeight,
			'$alpha' : 0,
			'$snap'   : false
		},
		'id':	'#DUMMY_GRID_NODE'
		}
	);
	
	
	  
	  fd = new $jit.GraphFlow({
		width : document.getElementById(visContainer).clientWidth,
		height : document.getElementById(visContainer).clientHeight,
	  
		//id of the visualization container
		injectInto: visContainer,
		//Enable zooming and panning
		//with scrolling and DnD
		Navigation: {
		  enable: true,
		  //Enable panning events only if we're dragging the empty
		  //canvas (and not a node).
		  panning: 'avoid nodes',
		  zooming: 40 //zoom speed. higher is more sensible
		},
		// Change node and edge styles such as
		// color and width.
		// These properties are also set per node
		// with dollar prefixed data-properties in the
		// JSON structure.
		Node: {
		  overridable: true,
		  CanvasStyles: {  
			  lineWidth: 2  
			}  
		  
		},
		Edge: {
		  overridable: true,
		  color: edgeColor,
		  lineWidth: 1,
		  dim: vertexLabelSize,
		  type: 'flowArrow'
		},
		//Native canvas text styling  
		Label: {  
			type: 'HTML', //Native or HTML  
			size: 10,  
			style: 'bold'  
		}, 
		//Add Tips  
		  Tips: {  
			enable: true,  
			type: 'Native',
			onShow: function(tip, node) {  
			  var message = node.data.$tooltip;
			  
			  // do not show a tooltip if it is undefined
			  if(message == undefined){
				tip.style.visibility = "hidden";
			  }
			  else{
				tip.innerHTML = "<div class=\"tip-text\">" + message + "</div>";  
				tip.style.visibility = "visible";
			  }
			}  
		 },  
		 
		// Add node events
		Events: {
		  enable: true,
		  enableForEdges: true,
		  type: 'Native',
		  
		  /**
		   *	EVENT:	OnRightClick
		   */
		  onRightClick: function(node, eventInfo, e) {
			callListener("onRightClick", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragCancel
		   */
		  onDragCancel: function(node, eventInfo, e) {
			callListener("onDragCancel", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseEnter
		   */
		  onMouseEnter: function(node, eventInfo, e) {
			if(node.id == mouseNode.id){
				return;
			}
			
			// ignore some highlights if we are dragging an edge
			if(selectedNode != null && 
				(node.data.$type == "nodeFamily" 
				|| node.data.$type == "crossBox")){
				return;
			}
			
			// check if object is an edge or a node
			mouseOverNode = (node.nodeFrom == undefined);
			mouseOverEdge = !mouseOverNode;
			
			setHighlight(node);
			if(readOnly.move && node.data.$type == "nodeFamily"){
				fd.canvas.getElement().style.cursor = 'move';
			}else if (readOnly.deleteCreate){
				fd.canvas.getElement().style.cursor = 'pointer';
			}
			callListener("onMouseEnter", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseMove
		   */
		  onMouseMove: function(node, eventInfo, e) {
			// are we dragging an edge with the mouse ?
			if(selectedNode){
				var pos = eventInfo.getPos();
				mouseNode.pos.setc(pos.x, pos.y-2);
				fd.plot();
			}
			callListener("onMouseMove", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseLeave
		   */
		  onMouseLeave: function(node, eventInfo, e) {
			fd.canvas.getElement().style.cursor = '';
			mouseOverNode = false;
			mouseOverEdge = false;
			
			setHighlight(null);
			callListener("onMouseLeave", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragStart
		   *	Actually this Event fires whenever the mouse button is down
		   */
		  onDragStart: function(node, eventInfo, e) {
		  var mousePos = eventInfo.getPos();
		  
		  // dragging the canvas
			if(!node || node.id == mouseNode.id ||
			   !node.data.$movable){
				
			    navi.dragX = e.layerX; //e.pageX;
				navi.dragY = e.layerY; //e.pageY;
				return;
			}
			
			//dragging a node
			if(!selectedNode){
				// memorize where we drag the node instead on centering
				// it to the mouse pointer
				var	nodePos = node.pos.getc(true);
				// the offset between the center of the node and the actual position
				// where we grabbed the node
				navi.dragX = nodePos.x - mousePos.x;
				navi.dragY = nodePos.y - mousePos.y;
				
				// is dragging a node
				navi.isDragging = true;
				
				/*
				// memorize the node position before node movement
				navi.dragX = nodePos.x;
				navi.dragY = nodePos.y;
				// prepare borders. These determine when to move the canvas with the node
				navi.borderLeft = (node.data.$width / 2 - navi.nodeOffsetX) * fd.canvas.scaleOffsetX;
				navi.borderTop = (node.data.$height / 2 - navi.nodeOffsetY) * fd.canvas.scaleOffsetX;
				navi.borderRight = grid.data.$width - (node.data.$width / 2 + navi.nodeOffsetX) * fd.canvas.scaleOffsetX;
				navi.borderBottom = grid.data.$height - (node.data.$height / 2 + navi.nodeOffsetY) * fd.canvas.scaleOffsetX;
				*/
			}
			callListener("onDragStart", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragMove
		   */
		  onDragMove: function(node, eventInfo, e) {
			if(!node || !node.data.$movable || selectedNode || !readOnly.move){
					return;
			}
			
			// move node
			var pos = eventInfo.getPos(),
				x = pos.x + navi.dragX,
				y = pos.y + navi.dragY;
			moveNode(node, x, y);
			
			callListener("onDragMove", [node, eventInfo, e]);
			fd.plot();
			
		  },
		  
		  /**
		   *	EVENT:	OnDragEnd
		   */
		  onDragEnd: function(node, eventInfo, e) {
			
			callListener("onDragEnd", [node, eventInfo, e]);
			
			if(node){
				var children = node.data.$children;
				var data, source, target;
				
				// this function recalculates the edge offset
				var updateOffset = function(n){
					n.eachAdjacency(function(adja){
						
						data = adja.data;
						
						// is the first node the source node?
						if(adja.nodeFrom.id == data.$direction[0]){
							source = adja.nodeFrom;
							target = adja.nodeTo;
						}
						else{
							source = adja.nodeTo;
							target = adja.nodeFrom;
						}
						data.$offsetFrom =
							customEdgeMap[source.data.$type].getOffset(source, target, true, data);
						data.$offsetTo =
							customEdgeMap[target.data.$type].getOffset(source, target, false, data);
					});
				}
				
				// recalculate the edge offset of the node and its children
				updateOffset(node);
				for(var c in children){
					updateOffset(children[c]);
				}
				
			}
			
			fd.plot();
		  },
		  
		  /**
		   *	EVENT:	OnTouchMove
		   */
		  onTouchMove: function(node, eventInfo, e) {
			$jit.util.event.stop(e); //stop default touchmove event
			this.onDragMove(node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnClick
		   */
		  onClick: function(node, eventInfo, e){
		  
		  var isNotDragging = true;
		  var wasDraggingEdge = navi.isDraggingEdge;
		  
		  // set mouse position (required for dragging edges)
		  var pos = eventInfo.getPos();
		  mouseNode.pos.setc(pos.x, pos.y-2);
		  
		  
		  // finished dragging the screen
		  if(!node){
			var deltaX = e.layerX - navi.dragX,
				deltaY = e.layerY - navi.dragY;
			navi.centerX -= deltaX;
			navi.centerY -= deltaY;
			
			isNotDragging = (Math.abs(deltaX) < 5 && Math.abs(deltaY) < 5);
			
			// remove selection and mouse-edge upon clicking anywhere
			if(selectedNode){
			  connectEdge(null);
			}
			fd.plot();
		  }
		  else if(!node.nodeFrom && readOnly.deleteCreate){
			
			var incomingEdge = false;
			switch(node.data.$type){
				case "crossBox":
					removeNode(node.data.$family, true);
					break;
					
				case "nodeFamily":
					break;
					
				case "inputPort":
					incomingEdge = true;
				case "outputPort":
				case "repositoryPort":
					connectEdge(node.id, incomingEdge);
					break;
			}
			
		  }
		  else if(node.nodeFrom && readOnly.deleteCreate){
			// no selection yet
				if(selectedNode == null){
					
					var source = node.data.$direction[0],
						label = node.data.$label;
					
					// remove selectedEdge
					removeEdge(source, node.data.$direction[1]);
					
					// create dangling edge
					connectEdge(source, false, label);
					
					fd.plot();
					
					return;
				}
		  }
		  
		  if(isNotDragging){
			  callListener("onClick", [node, eventInfo, e]);
		  }
		  
		  navi.isDragging = false;
		  
		  if(wasDraggingEdge){
			  navi.isDraggingEdge = false;
		  }
		  
		  },
		  
		  /**
		   * 	EVENT: OnMouseWheel
		   */
		  onMouseWheel: function(e, delta) {
			  // update whether labels should be displayed
			  fd.canvas.showLabels = (fd.canvas.scaleOffsetX > fd.canvas.labelThreshold);
			  fd.plot();
		  }
		},
		
		
		//Number of iterations for the FD algorithm
		iterations: 0,
		//Edge length
		levelDistance: 0
	  });
	  
	  // load JSON data.
	  fd.loadJSON(json, 0);
	  
	  // set pointers to grid and mouse node
	  mouseNode = fd.graph.getNode("#DUMMY_MOUSE_NODE");
	  grid = fd.graph.getNode("#DUMMY_GRID_NODE");
	  
	  // set zoom threshold for displaying text
	  fd.canvas.labelThreshold = 1/vertexLabelSize;
	  fd.canvas.showLabels = true;
	  
	  console.log(fd.canvas);
	  console.log(fd.graph);
	}
	
	// initialize graph
	initGraph();
	return this;

}
 